/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdarg>
#include <sqlite3.h>

#include "database.hpp"
#include "event.hpp"

//TODO: http://www.sqlite.org/compile.html

struct QueryObject
{
	QueryObject()
	{
		refcount = 0;
		db = 0;
	}
	
	~QueryObject()
	{
		if (statement){
			sqlite3_finalize(statement);
		}
	}
	
	void DecRefCount()
	{
		refcount--;
		if (!refcount){
			delete this;
		}
	}
	
	std::string sql;
	sqlite3 *db;
	sqlite3_stmt *statement;
	unsigned int refcount;
	
	//a pointer to Database *parent maybe?
};

Query::Query()
{
	o = 0;
}

Query::Query(QueryObject *po)
{
	o = po;
	if (o){
		o->refcount++;
	}
}

Query::~Query()
{
	if (o){
		o->DecRefCount();
	}
}

Query::Query(const Query &rhs)
{
	o = rhs.o;
	if (o){
		o->refcount++;
	}
}

Query &Query::operator=(const Query &rhs)
{
	if (this != &rhs){
		if (o){
			o->DecRefCount();
		}
		o = rhs.o;
		o->refcount++;
	}
	return *this;
}

DynamicValue Query::operator()()
{
	DynamicValue v;
	return (*this)(v);
}

DynamicValue Query::operator()(DynamicValue v)
{
	//So we're going to use the name from the statement to look up in the dictionary
	//Therefore we expect that:
	//a) Named values are used
	//b) v is a map
	//In the case of dumping stuff from an event, it might be appropriate to simply
	//use the event its self as v
	
	if (!o->statement){
		throw FormatError("Cannot execute uninitialised query");
	}
	
	unsigned int n = sqlite3_bind_parameter_count(o->statement) + 1;
	for (unsigned int i = 1; i < n; i++){
		const char *name = sqlite3_bind_parameter_name(o->statement, i) + 1;
		if (!name){
			throw FormatError("Query %s has unnamed parameter for %u/%u", o->sql.c_str(), i, n - 1);
		}
		if (!v.Has(name)){
			throw FormatError("Query map does not have %s", name);
		}
		
		//TODO: check that this gives the desired result
		if (v[name].Type<bool>()){
			sqlite3_bind_int(o->statement, i, v[name]);
		}
		else if (v[name].Type<int64_t>()){
			sqlite3_bind_int64(o->statement, i, (int64_t)v[name]);
		}
		else if (v[name].Type<double>()){
			sqlite3_bind_double(o->statement, i, v[name]);
		}
		else if (v[name].Type<std::string>()){
			sqlite3_bind_text(o->statement, i, std::string(v[name]).c_str(), -1, SQLITE_STATIC);
		}
		else if (v[name].Type<Buffer>()){
			Buffer b = v[name];
			sqlite3_bind_blob(o->statement, i, &b[0], b.GetLength(), SQLITE_STATIC);
		}
		else {
			throw FormatError("Value for %s could not be converted for SQL: %s",
				name, v[name].ToString().c_str());
		}
	}
	
	std::vector<DynamicValue> table;
	while (true){
		int e = sqlite3_step(o->statement);
		if (SQLITE_DONE == e){
			if (sqlite3_reset(o->statement) != SQLITE_OK){
				throw FormatError("Could not reset query");
			}
			break;
		}
		else if (SQLITE_ERROR == e){
			throw FormatError("Could not execute query %s: %s", o->sql.c_str(), sqlite3_errmsg(o->db));
		}
		else if (SQLITE_ROW == e){
			DynamicValue row;
			const void *blobv = 0;
			int blobn = 0;
			
			int n = sqlite3_column_count(o->statement);
			for (int i = 0; i < n; i++){
				int type = sqlite3_column_type(o->statement, i);
				DynamicValue col;
				switch (type)
				{
					case SQLITE_INTEGER:
						col = (int64_t)sqlite3_column_int64(o->statement, i);
						break;
					case SQLITE_FLOAT:
						col = sqlite3_column_double(o->statement, i);
						break;
					case SQLITE_TEXT:
						col = std::string((const char *)sqlite3_column_text(o->statement, i));
						break;
					case SQLITE_BLOB:
						blobv = sqlite3_column_blob(o->statement, i);
						blobn = sqlite3_column_bytes(o->statement, i);
						col = Buffer((const unsigned char *)blobv, blobn);
						break;
					case SQLITE_NULL:
						break;
					default:
						throw FormatError("Unknown column type");
				}
				const char *name = sqlite3_column_name(o->statement, i);
				if (!name){
					throw FormatError("Unknown column name");
				}
				row[name] = col;
			}
			table.push_back(row);
		}
		else {
			throw FormatError("An unknown condition occured: %i (SQLITE_DONE = %i, SQLITE_ERROR = %i, SQLITE_ROW = %i)", e, SQLITE_ERROR, SQLITE_ROW);
		}
	}
	return table;
}

Database::Database(Configuration &pconfiguration) :
	configuration(pconfiguration)
{
	std::string filename = configuration["database"]["filename"];
	sqlite3_open(filename.c_str(), &db);
}

Database::~Database()
{
	sqlite3_close(db);
}

void Database::Init(EventManager *pmanager)
{
	manager = pmanager;
}

Query Database::operator()(std::string sql)
{
	QueryObject *o = new QueryObject;
	o->sql = sql;
	o->db = db;
	
	//TODO learn how to use the last argument to discover if there was only one statement
	int e = sqlite3_prepare_v2(db, sql.c_str(), -1, &o->statement, 0);
	if (e != SQLITE_OK){
		throw FormatError("Could not prepare statement: %s: %s", sql.c_str(), sqlite3_errmsg(db));
	}
	
	return Query(o);
}
