/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>
#include "dynamic.hpp"
#include "format.hpp"

Format::Format()
{
	refcount = 0;
}

DynamicValueFormat::DynamicValueFormat(Format *f)
{
	format = f;
	if (format){
		format->refcount++;
	}
}

DynamicValueFormat::DynamicValueFormat(const DynamicValueFormat &rhs)
{
	format = rhs.format;
	if (format){
		format->refcount++;
	}
}

DynamicValueFormat &DynamicValueFormat::operator=(const DynamicValueFormat &rhs)
{
	if (this != &rhs){
		if (format){
			format->refcount--;
			if (!format->refcount){
				delete format;
			}
		}
		
		format = rhs.format;
		if (format){
			format->refcount++;
		}
	}
	return *this;
}

DynamicValueFormat::~DynamicValueFormat()
{
	if (format){
		format->refcount--;
		if (!format->refcount){
			delete format;
		}
	}
}

DynamicValueFormat::operator bool() const
{
	return format;
}

std::string DynamicValueFormat::DoFormat(DynamicValue &v)
{
	if (format){
		return format->DoFormat(v);
	}
	return "";
}



class Ipv4Format : public Format
{
	public:
		virtual std::string DoFormat(DynamicValue &v)
		{
			try {
				char c[16];
				Buffer b = v;
				if (b.GetLength() != 4){
					return "Invalid IPv4 Address";
				}
				snprintf(c, 16, "%u.%u.%u.%u", b[0], b[1], b[2], b[3]);
				return c;
			}
			catch (Error e) {
				return "No IPv4 Address";
			}
		}
};

DynamicValueFormat FormatIpv4()
{
	return DynamicValueFormat(new Ipv4Format());
}

class MacFormat : public Format
{
	public:
		virtual std::string DoFormat(DynamicValue &v)
		{
			try {
				char c[18];
				Buffer b = v;
				if (b.GetLength() != 6){
					return "Invalid MAC Address";
				}
				snprintf(c, 18, "%02X:%02X:%02X:%02X:%02X:%02X",
					b[0], b[1], b[2], b[3], b[4], b[5]);
				return c;
			}
			catch (Error e) {
				return "No MAC Address";
			}
		}
};

DynamicValueFormat FormatMac()
{
	return DynamicValueFormat(new MacFormat());
}

class HexFormat : public Format
{
	public:
		HexFormat(unsigned int pn)
		{
			n = pn;
		}
		
		virtual std::string DoFormat(DynamicValue &v)
		{
			try {
				char c[32];
				unsigned int i = v;
				
				if (n){
					char f[32];
					snprintf(f, 32, "0x%%0%uX", n);
					snprintf(c, 32, f, i);
				}
				else {
					snprintf(c, 32, "0x%X", i);
				}
				return c;
			}
			catch (Error e) {
				return "No MAC Address";
			}
		}
		
	private:
		unsigned int n;
};

DynamicValueFormat FormatHex(unsigned int n)
{
	return DynamicValueFormat(new HexFormat(n));
}
