/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011-2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include "event.hpp"
#include "modules.hpp"

#ifdef MODULES
	#include <cstring>
	#include <dirent.h>
	#include <dlfcn.h>
	
	#include <string>
	
	int dlopenflags = RTLD_LAZY | RTLD_GLOBAL;
	
	void Modules::LoadModule(EventManager *manager, const char *path)
	{
		void* handle = dlopen(path, dlopenflags);
		
		if (!handle){
			throw FormatError("Could not open %s: %s", path, dlerror());
		}
		
		Module *(*InitModule)(Modules *, EventManager *, Configuration *, Database *) =
			(Module *(*)(Modules *, EventManager *, Configuration *, Database *))dlsym(handle, "InitModule");
		
		if (!InitModule){
			throw FormatError("GetInitModule() returned zero");
		}
		
		AddModule(InitModule(this, manager, configuration, database), path, handle);
	}
	
	void Modules::InitModules(EventManager *manager, const char *path)
	{
		if (!path){
			std::vector<DynamicValue> l = (*configuration)["hyperion"]["modules"];
			for (std::vector<DynamicValue>::iterator it = l.begin(); it != l.end(); it++){
				std::string str = *it;
				//TODO: Improve this by making it see if str is a directory or a file
				//Or even use file to see if it's a module or not
				//(Or even better - allow the use of a regex)
				if (str[str.size()-1] == '/'){
					InitModules(manager, str.c_str());
				}
				else {
					LoadModule(manager, str.c_str());
				}
			}
			return;
		}
		
		const char *ModuleExtension = ((std::string)((*configuration)["hyperion"]["module_extension"])).c_str();
		unsigned int ModuleExtension_length = strlen(ModuleExtension);
		
		//for each so in the directory
		DIR *directory = opendir(path);
		if (!directory){
			throw FormatError("Could not open directory \"%s\"", path);
		}
		
		while (true){
			struct dirent *entry = readdir(directory);
			if (!entry){
				break;
			}
			
			const char *filename = entry->d_name;
			unsigned int filename_length = strlen(filename);
			if (filename_length < ModuleExtension_length){
				continue;
			}
			if (strcmp(filename + filename_length - ModuleExtension_length, ModuleExtension)){
				continue;
			}
			std::string sopath = std::string(path) + std::string(filename);
			LoadModule(manager, sopath.c_str());
		}
		
		closedir(directory);
	}

#else
	#define RegisterModule(moduleclass) \
		Module *InitModule_##moduleclass(Modules *, EventManager *, Configuration *, Database *); \
		AddModule(InitModule_##moduleclass(this, manager, configuration, database));
	
	void Modules::InitModules(EventManager *manager, const char *path)
	{
		if (!path){
			#include "modules/modules_static.hpp"
		}
		else {
			throw FormatError("Module support not enables, so could not load modules from \"%s\"");
		}
	}
	
	void Modules::LoadModule(EventManager *manager, const char *path)
	{
		throw FormatError("Module support not enabled, so could not load module \"%s\"");
	}
#endif

Modules::Modules(Configuration *pconfiguration, Database *pdatabase)
{
	configuration = pconfiguration;
	database = pdatabase;
}

Modules::~Modules()
{
	for (std::map<std::string, ModulesItem>::iterator it = modules.begin(); it != modules.end(); it++){
		delete it->second.module;
		#ifdef MODULES
		if (it->second.handle){
			dlclose(it->second.handle);
		}
		#endif
	}
}

void Modules::AddModule(Module *module)
{
	AddModule(module, 0, 0);
}

void Modules::AddModule(Module *module, const char *path, void *handle)
{
	std::string name = module->GetName();
	if (modules.count(name)){
		throw FormatError("Module \"%s\" has already been loaded", module->GetName());
	}
	
	modules[name].module = module;
	if (path){
		modules[name].path = path;
	}
	modules[name].handle = handle;
	
	if ((*configuration)["hyperion"]["debug"]){
		printf("Added module \"%s\"\n", module->GetName());
	}
}

Module *Modules::GetModule(std::string name)
{
	if (modules.count(name)){
		return modules[name].module;
	}
	return 0;
}
