/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>
#include <typeinfo>

#include <pcrecpp.h>

#include "event.hpp"
#include "module.hpp"

static Buffer GetIpv4AddressFromString(std::string str)
{
	unsigned int octs[4];
	if (sscanf(str.c_str(), "%u.%u.%u.%u", &octs[0], &octs[1], &octs[2], &octs[3]) != 4){
		throw FormatError("Could not convert %s to IP address", str.c_str());
	}
	
	Buffer b(4);
	b.SetUInt8(0, octs[0]);
	b.SetUInt8(1, octs[1]);
	b.SetUInt8(2, octs[2]);
	b.SetUInt8(3, octs[3]);
	
	return b;
}

static Buffer GetMacAddressFromString(std::string str)
{
	unsigned int octs[6];
	if (sscanf(str.c_str(), "%02X:%02X:%02X:%02X:%02X:%02X",
		&octs[0], &octs[1], &octs[2], &octs[3], &octs[4], &octs[5]) != 6){
		throw FormatError("Could not convert %s to MAC address", str.c_str());
	}
	
	Buffer b(6);
	b.SetUInt8(0, octs[0]);
	b.SetUInt8(1, octs[1]);
	b.SetUInt8(2, octs[2]);
	b.SetUInt8(3, octs[3]);
	b.SetUInt8(4, octs[4]);
	b.SetUInt8(5, octs[5]);
	
	return b;
}



//TODO: come up with a way to make this work for multihomed devices?
//This probably means making changes to how modules are layed out (making them a tree perhaps?)
//Rather than making changes here

class Route
{
	public:
		Route(std::string src, std::string dst, std::string strip) :
			src_regex(src), dst_regex(dst)
		{
			if (strip != ""){
				ip = GetIpv4AddressFromString(strip);
			}
			else {
				ip = Buffer();
			}
		}
		
		bool Match(DynamicValue src, DynamicValue dst)
		{
			return (src_regex.FullMatch(src.ToString()) && dst_regex.FullMatch(dst.ToString()));
		}
		
		DynamicValue GetIp()
		{
			return ip;
		}
	
	private:
		pcrecpp::RE src_regex;
		pcrecpp::RE dst_regex;
		DynamicValue ip;
};

class ArpModule : public Module
{
	public:
		ArpModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Arp", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<ArpModule>
				("received:arp", this, &ArpModule::OnArp));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<ArpModule>
				("send:auto:ipv4", this, &ArpModule::OnSendAutoIpv4));
			
			//MAC, IP, Sources regex, Destinations regex
			std::vector<DynamicValue> croutes = GetConfiguration()["arp"]["routes"];
			for (std::vector<DynamicValue>::iterator it = croutes.begin(); it != croutes.end(); it++){
				std::vector<DynamicValue> croute = *it;
				
				if ((std::string(croute[0]) != "") && (std::string(croute[1]) != "")){
					cache[GetIpv4AddressFromString(croute[1])] = GetMacAddressFromString(croute[0]);
				}
				
				Route route(croute[2], croute[3], croute[1]);
				routes.push_back(route);
			}
			
			std::vector<DynamicValue> ips = GetConfiguration()["arp"]["ips"];
			for (std::vector<DynamicValue>::iterator it = ips.begin(); it != ips.end(); it++){
				std::vector<DynamicValue> ip = *it;
				local[GetIpv4AddressFromString(ip[1])] = GetMacAddressFromString(ip[0]);
			}
		}
		
	private:
		void OnSendAutoIpv4(Event event)
		{
			//TODO: Have a prev-name
			Event prev = event["send:auto:ipv4"]["previous"];
			DynamicValue source = prev["send:ipv4"]["source"];
			DynamicValue destination = prev["send:ipv4"]["destination"];
			
			for (std::list<Route>::iterator it = routes.begin(); it != routes.end(); it++){
				if (it->Match(source.ToString(), destination.ToString())){
					//We found the route, let's use it!
					DynamicValue ip = it->GetIp();
					Buffer bip = ip;
					if (!bip.GetLength()){
						ip = destination;
					}
					
					if (!cache.count(ip)){
						/*throw FormatError*/printf("Cannot send to %s because it's not in the cache\n",
							ip.ToString().c_str());
						return;
					}
					
					if (!local.count(source)){
						/*throw FormatError*/printf("Cannot find MAC address for source %s\n", source.ToString().c_str());
						return;
					}
					
					Event e;
					Event sniff;
					sniff["send:sniff"]["device"] = ""; //TODO
					e["send:ethernet"]["source"] = local[source];
					e["send:ethernet"]["destination"] = cache[ip];
					e["send:ethernet"]["source"].Format(FormatMac());
					e["send:ethernet"]["destination"].Format(FormatMac());
					e["send:ethernet"]["protocol"] = 0x0800;
					e["send:ethernet"]["previous"] = prev;
					e["send:ethernet"]["next"] = sniff;
					e["send:ethernet"]["next-name"] = "send:sniff";
					e["send:ethernet"]["payload"] = event["send:auto:ipv4"]["payload"];
					GetManager()->Publish(e);
					return;
				}
			}
			/*throw FormatError*/printf("Route for %s to %s is undefined\n", source.ToString().c_str(), destination.ToString().c_str());
		}
		
		void OnArp(Event event)
		{
			bool anomaly = false;
			Buffer b = event["received:arp"]["data"];
			if (b.GetLength() < 28){
				Event e;
				event["anomaly:received:arp:incomplete arp"]["data"] = b;
				GetManager()->Publish(e);
				return;
			}
			
			unsigned int hwtype = b.GetUInt16N(0);
			unsigned int prototype = b.GetUInt16N(2);
			unsigned int hwlen = b.GetUInt8(4);
			unsigned int protolen = b.GetUInt8(5);
			unsigned int op = b.GetUInt16N(6);
			
			Buffer senderhw = Buffer(b, 8, 6);
			Buffer senderproto = Buffer(b, 14, 4);
			Buffer targethw = Buffer(b, 18, 6);
			Buffer targetproto = Buffer(b, 24, 4);
			
			DynamicValue v;
			v["hardware type"] = hwtype;
			v["protocol type"] = prototype;
			v["hardware length"] = hwlen;
			v["protocol length"] = protolen;
			v["operation"] = op;
			//v["string operation"] = (op == 1) ? "request" : (op == 2) ? "reply" : "unknown";
			v["sender hardare"] = senderhw;
			v["sender protocol"] = senderproto;
			v["target hardware"] = targethw;
			v["target protocol"] = targetproto;
			
			v["sender hardare"].Format(FormatMac());
			v["sender protocol"].Format(FormatIpv4());
			v["target hardware"].Format(FormatMac());
			v["target protocol"].Format(FormatIpv4());
			
			if (hwtype != 1){
				Event e;
				v["data"] = b;
				e["anomaly:received:arp:hardware type"] = v;
				GetManager()->Publish(e);
				anomaly = true;
			}
			
			if (prototype != 0x0800){
				Event e;
				v["data"] = b;
				e["anomaly:received:arp:protocol type"] = v;
				GetManager()->Publish(e);
				anomaly = true;
			}
			
			if (hwlen != 6){
				Event e;
				v["data"] = b;
				e["anomaly:received:arp:hardware length"] = v;
				GetManager()->Publish(e);
				anomaly = true;
			}
			
			if (protolen != 4){
				Event e;
				v["data"] = b;
				e["anomaly:received:arp:protocol length"] = v;
				GetManager()->Publish(e);
				anomaly = true;
			}
			
			if ((op != 1) && (op != 2)){
				Event e;
				v["data"] = b;
				e["anomaly:received:arp:operation"] = v;
				GetManager()->Publish(e);
				anomaly = true;
			}
			
			if (anomaly){
				return;
			}
			
			if (cache.count(senderproto)){
				if (cache[senderproto] != senderhw){
					Event e;
					v["current"] = cache[senderproto];
					v["new"] = senderhw;
					e["anomaly:received:arp:alters cache"] = v;
					GetManager()->Publish(e);
					return;
				}
			}
			else {
				cache[senderproto] = senderhw;
				Event e;
				e["arp:new"]["ipv4"] = senderproto;
				e["arp:new"]["ethernet"] = senderhw;
				GetManager()->Publish(e);
			}
			
			Event e;
			e["processed:arp"] = v;
			GetManager()->Publish(e);
		}
		
		std::map<Buffer, Buffer> cache;
		std::map<Buffer, Buffer> local;
		std::list<Route> routes;
};

INIT_MODULE(ArpModule);
