/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>
#include <typeinfo>

#include "event.hpp"
#include "module.hpp"

class IcmpModule : public Module
{
	public:
		IcmpModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("ICMP", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<IcmpModule>
				("received:icmp", this, &IcmpModule::OnReceive));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<IcmpModule>
				("send:icmp:echo request:echo request", this, &IcmpModule::OnSendEchoRequest));
		}
		
	private:
		void OnReceive(Event event)
		{
			Event e;
			Buffer b = event["received:icmp"]["data"];
			if (b.GetLength() < 4){
				e["anomaly:received:icmp:incomplete"];
				GetManager()->Publish(e);
				return;
			}
			
			unsigned int type = b.GetUInt8(0);
			unsigned int code = b.GetUInt8(1);
			//unsigned int checksum = event.GetBuffer().GetUInt16N(2);
			
			b = Buffer(b, 4);
			switch (type)
			{
				case 0:
					EchoReply(type, code, b);
					break;
				case 8:
					EchoRequest(type, code, b);
					break;
				case 11:
					TimeExceeded(type, code, b);
					break;
				default:
					e["anomaly:received:icmp:unknown"]["type"] = type;
					e["anomaly:received:icmp:unknown"]["code"] = code;
					e["anomaly:received:icmp:unknown"]["payload"] = b;
					GetManager()->Publish(e);
			}
		}
		
		void EchoRequest(unsigned int type, unsigned int code, Buffer buffer)
		{
			Event e;
			if (!code){ //0 is the only valid code for this
				e["processed:icmp:echo request:echo request"]["id"] = buffer.GetUInt16N(0);
				e["processed:icmp:echo request:echo request"]["seq"] = buffer.GetUInt16N(2);
				e["processed:icmp:echo request:echo request"]["payload"] = Buffer(buffer, 4);
				GetManager()->Publish(e);
				return;
			}
			e["anomaly:received:icmp:request:unknown"]["code"] = code;
			e["anomaly:received:icmp:request:unknown"]["payload"] = buffer;
			GetManager()->Publish(e);
		}
		
		void EchoReply(unsigned int type, unsigned int code, Buffer buffer)
		{
			Event e;
			if (!code){ //0 is the only valid code for this
				e["processed:icmp:echo reply:echo reply"]["id"] = buffer.GetUInt16N(0);
				e["processed:icmp:echo reply:echo reply"]["seq"] = buffer.GetUInt16N(2);
				e["processed:icmp:echo reply:echo reply"]["payload"] = Buffer(buffer, 4);
				GetManager()->Publish(e);
				return;
			}
			e["anomaly:received:icmp:echo reply:unknown"]["code"] = code;
			e["anomaly:received:icmp:echo reply:unknown"]["payload"] = buffer;
			GetManager()->Publish(e);
		}
		
		void TimeExceeded(unsigned int type, unsigned int code, Buffer buffer)
		{
			Event e;
			if (buffer.GetLength() < 24){
				e["anomaly:received:icmp:time exceeded:length"]["buffer"] = buffer;
				GetManager()->Publish(e);
			}
			
			unsigned int headerlength = buffer.GetUInt8(4) & 0x0F;
			headerlength *= 4;
			switch (code)
			{
				case 0:
					if (headerlength < 20){
						e["anomaly:received:icmp:time exceeded:ttl:header length"]["length"] = headerlength;
						e["anomaly:received:icmp:time exceeded:ttl:header length"]["buffer"] = buffer;
						break;
					}
					if (buffer.GetLength() < headerlength){
						e["anomaly:received:icmp:time exceeded:ttl:length"]["buffer"] = buffer;
						break;
					}
					
					e["processed:icmp:time exceeded:ttl"]["protocol"] = buffer[13];
					e["processed:icmp:time exceeded:ttl"]["source"] = Buffer(buffer, 16, 4);
					e["processed:icmp:time exceeded:ttl"]["source"].Format(FormatIpv4());
					e["processed:icmp:time exceeded:ttl"]["destination"] = Buffer(buffer, 20, 4);
					e["processed:icmp:time exceeded:ttl"]["destination"].Format(FormatIpv4());
					e["processed:icmp:time exceeded:ttl"]["payload"] = Buffer(buffer, 4 + headerlength);
					break;
				default:
					e["anomaly:received:icmp:time exceeded:unknown"]["code"] = code;
					e["anomaly:received:icmp:time exceeded:unknown"]["payload"] = buffer;
			}
			GetManager()->Publish(e);
		}
		
		void OnSendEchoRequest(Event event)
		{
			Buffer b(8);
			b.SetUInt8(0, 8);
			b.SetUInt8(1, 0);
			b.SetUInt16N(4, event["send:icmp:echo request:echo request"]["id"]);
			b.SetUInt16N(6, event["send:icmp:echo request:echo request"]["seq"]);
			b += event["send:icmp:echo request:echo request"]["payload"];
			b.Rfc1071(2);
			
			std::string name = event["send:icmp:echo request:echo request"]["next-name"];
			Event e = event["send:icmp:echo request:echo request"]["next"];
			e[name]["payload"] = b;
			e[name]["previous"] = event;
			GetManager()->Publish(e);
		}
};

INIT_MODULE(IcmpModule);
