/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

//TODO: Fragments

#include <cstdio>
#include <typeinfo>

#include "event.hpp"
#include "module.hpp"

class Ipv4Module : public Module
{
	public:
		Ipv4Module(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("IPv4", modules, manager, configuration, database)
		{
			//See? No problem if we register twice! The second one just gets ignored
			GetManager()->GetIntEventTable("ethernet")->Add("received:ipv4", 0x0800);
			
			//IP, rather than IPv4 - because this will enable IPv6 to share the same table
			//For IPv6, while it's OK to have an event for, e.g. "Destination Options for IPv6",
			//The IPv6 processor should process them inline
			//This means:
			//a) We don't have to worry about IPv6 processing bad IPv4 packets
			//b) We can use it purely for an anomaly where IPv4 has a protocol
			//	 of an IPv6 header
			//GetManager()->GetIntEventTable("IP")->Add<UdpReceiveEvent>("UDP", 17);
			
			
			//we should probably define a standard for where these are kept in the default
			//case (even if it's possibe to put it in either)
			GetManager()->GetIntEventTable("ip")->Add("received:icmp", 1);
			GetManager()->GetIntEventTable("ip")->Add("received:tcp", 6);
			GetManager()->GetIntEventTable("ip")->Add("received:udp", 17);
			
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<Ipv4Module>
				("received:ipv4", this, &Ipv4Module::OnReceive));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<Ipv4Module>
				("send:ipv4", this, &Ipv4Module::OnSend));
		}
		
	private:
		void OnReceive(Event event)
		{
			//TODO: Implement more validation (e.g. checksum) (also in other protocols)
			//TODO: make these errors include more dynamic information
			
			Buffer b = event["received:ipv4"]["data"];
			if (b.GetLength() < 20){
				Event e;
				e["anomaly:received:ipv4:incomplete header"]["data"] = b;
				GetManager()->Publish(e);
				return;
			}
			
			
			uint8_t proto = b.GetUInt8(9);
			unsigned int headerlength = b.GetUInt8(0) & 0x0F;
			unsigned int packetlength = b.GetUInt16N(2);
			//unsigned int version = (event->GetBuffer().GetUInt8(0) & 0xF0) >> 4;
			headerlength *= 4; //in fact, the length is represented in the packet
				//as a multiple of 4 bytes
			
			bool processable = true;
			if (packetlength < headerlength){
				Event e;
				e["anomaly:received:ipv4:packet length < header length"]["header length"] = headerlength;
				e["anomaly:received:ipv4:packet length < header length"]["packet length"] = packetlength;
				e["anomaly:received:ipv4:packet length < header length"]["data"] = b;
				GetManager()->Publish(e);
				processable = false;
			}
			if (b.GetLength() < packetlength){
				Event e;
				e["anomaly:received:ipv4:incomplete packet"]["packet length"] = packetlength;
				e["anomaly:received:ipv4:incomplete packet"]["data"] = b;
				GetManager()->Publish(e);
				processable = false;
			}
			
			if (!processable){
				return;
			}
			
			//See below
			Event e;
			unsigned int length = packetlength - headerlength;
			Buffer payload = Buffer(b, headerlength, length);
			
			e["processed:ipv4"]["payload"] = payload;
			e["processed:ipv4"]["protocol"] = proto;
			e["processed:ipv4"]["source"] = Buffer(&b[12], 4);
			e["processed:ipv4"]["destination"] = Buffer(&b[16], 4);
			e["processed:ipv4"]["source"].Format(FormatIpv4());
			e["processed:ipv4"]["destination"].Format(FormatIpv4());
			
			/*if ((e["processed:ipv4"]["source"].ToString() != "192.168.1.55") && (e["processed:ipv4"]["destination"].ToString() != "192.168.1.55")){
				return;
			}*/
			
			GetManager()->Publish(e);
			
			//maybe make this check that the IPv4 packet came to/from the right MAC address etc
			GetManager()->GetIntEventTable("ip")->Publish(proto, payload);
		}
		
		void OnSend(Event event)
		{
			Buffer payload = event["send:ipv4"]["payload"];
			Buffer b(12);
			b.SetUInt8(0, 0x45);
			b.SetUInt8(1, 0);
			b.SetUInt16N(2, 20 + payload.GetLength());
			b.SetUInt32(4, 0);
			
			if (event["send:ipv4"].Has("ttl")){
				b.SetUInt8(8, event["send:ipv4"]["ttl"]);
			}
			else {
				b.SetUInt8(8, 255);
			}
			b.SetUInt8(9, event["send:ipv4"]["protocol"]);
			b += event["send:ipv4"]["source"];
			b += event["send:ipv4"]["destination"];
			b.Rfc1071(10); //We don't include the payload for this checksum
			b += payload;
			
			std::string name = event["send:ipv4"]["next-name"];
			Event e = event["send:ipv4"]["next"];
			e[name]["payload"] = b;
			e[name]["previous"] = event;
			GetManager()->Publish(e);
		}
		
		//For sending: have it such that things like ttl can be set, but that if the event
		//does not include this, then it's set to a default
		
		std::string GetAddress(const unsigned char addr[4]) const
		{
			char buffer[16];
			snprintf(buffer, 16, "%u.%u.%u.%u",
				addr[0], addr[1], addr[2], addr[3]);
			return std::string(buffer);
		}
};

INIT_MODULE(Ipv4Module);
