/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>
#include <typeinfo>

#include <pcrecpp.h>

#include "module.hpp"

//TODO: Implement something that will take an event and guess at the event stack that produced it
//Maybe gather all events matching pattern 1 between sniff events in a list stored here, and
//display if any of the events (including the ones not matched by pattern 1) match pattern 2
//(or alternatively, perform the matching on pattern 1 if any packet matches pattern 2 - less
//processing of pattern 1 this way (i.e. for anomalies, not everything))

class LoggerModule : public Module
{
	public:
		LoggerModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Logger", modules, manager, configuration, database),
			filter(GetConfiguration()["logger"]["filter_regex"]),
			filter_exception(GetConfiguration()["logger"]["filter_exception_regex"]),
			filter_stack(GetConfiguration()["logger"]["filter_stack_regex"]),
			filter_stack_exception(GetConfiguration()["logger"]["filter_stack_exception_regex"]),
			filter_stack_trigger(GetConfiguration()["logger"]["filter_stack_trigger_regex"]),
			filter_stack_trigger_exception(GetConfiguration()["logger"]["filter_stack_trigger_exception_regex"])
		{
			GetManager()->Subscribe(new CallbackEventSubscription<LoggerModule>(this, &LoggerModule::OnEvent));
			
			//Set up valuefilters
			std::vector<DynamicValue> fsvrs = GetConfiguration()["logger"]["filter_stack_value_regexs"];
			for (std::vector<DynamicValue>::iterator it = fsvrs.begin(); it != fsvrs.end(); it++){
				std::list<std::string> fsvri;
				std::vector<DynamicValue> fsvr = *it;
				for (std::vector<DynamicValue>::iterator jt = fsvr.begin(); jt != fsvr.end(); jt++){
					fsvri.push_back(std::string(*jt));
				}
				valuefilters.push_back(fsvri);
			}
		}
		
	private:
		void OnEvent(Event event)
		{
			if (GetConfiguration()["logger"]["stack"]){
				if (event.Has("sniff")){
					std::vector<DynamicValue> v;
					v.reserve(events.size()); //optimisation
					
					unsigned int ntriggers = 0;
					unsigned int nfilters = 0;
					
					for (std::list<Event>::iterator it = events.begin(); it != events.end(); it++){
						if (
							!GetConfiguration()["logger"]["filter_stack"] ||
							(!filter_stack.FullMatch(it->GetFilter())) ||
							(filter_stack_exception.FullMatch(it->GetFilter()))){
							v.push_back(*it);
						}
						
						if (
							!GetConfiguration()["logger"]["filter_stack_trigger"] ||
							(!filter_stack_trigger.FullMatch(it->GetFilter())) ||
							(filter_stack_trigger_exception.FullMatch(it->GetFilter()))){
							ntriggers++;
						}
					}
					
					//TODO: Update this so if there are more unfiltered anomalous packets than there
					//are matching value filters, we don't block it
					if (ntriggers){ //we only really want to do the detailed
						//investigation if we need to
						
						//Look in each of the events to see if there's a match
						for (std::list<Event>::iterator it = events.begin(); it != events.end(); it++){
							//Iterate over the value filters
							for (std::list<std::list<std::string> >::iterator jt = valuefilters.begin();
								jt != valuefilters.end(); jt++){
								//Now we're in a specific value filter
								std::list<std::string>::iterator kt = jt->begin(); //kt is the current
									//string from the filter (first at the moment)
								
								//See if there's an event matching the first string in the filter
								if (!it->Has(*kt)){ //This event in the list does not have this event from
									//the filter
									//Try the next filter
									continue; //this does jt++, right?
								}
								//printf("Yay:D %s has %s :D\n", it->GetFilter().c_str(), kt->c_str());
								
								//Get the dynamic value that represents the event
								//(or later the part of the event)
								DynamicValue v = (*it)[*kt];
								kt++; //Make kt the next string
								
								bool found = false;
								while (kt != jt->end()){
									std::string str = *kt; //current string
									kt++; //get the next string
									//(this is done here in order to compare
									//to the end later)
									if (kt == jt->end()){ //this is the last one, so we compare on the
										//regex rather than the string its self
										pcrecpp::RE re(str); //construct the regex
										if (re.FullMatch(std::string(v.ToString()))){
											//we should surpress this packet
											nfilters++;
											found = true;
											break; //there's no need to keep looking
										}
									}
									else {
										if (!v.Has(str)){
											break; //we shouldn't kee decending
										}
										v = v[str]; //make v the next item
									}
								}
								//If the same event matches more than one filter, we consider this to be
								//one match in our tally
								if (found){ //there's no point to keep iterating here either
									break;
								}
							}
							//but we still need to look for other packets that might match
						}
					}
					
					if (ntriggers && (ntriggers > nfilters)){
						Event e;
						e["logger:stack"] = v;
						GetManager()->Publish(e);
					}
					events.clear();
				}
				if (!event.Has("logger:stack")){ //we probably want this
					events.push_back(event);
				}
			}
			
			if (GetConfiguration()["logger"]["print"]){
				if (
					!GetConfiguration()["logger"]["filter"] ||
					(!filter.FullMatch(event.GetFilter())) ||
					(filter_exception.FullMatch(event.GetFilter()))){
					printf("[*] %s: %s\n", event.GetFilter().c_str(), event.ToString().c_str());
				}
			}
		}
		
		std::list<Event> events;
		
		pcrecpp::RE filter;
		pcrecpp::RE filter_exception;
		pcrecpp::RE filter_stack;
		pcrecpp::RE filter_stack_exception;
		pcrecpp::RE filter_stack_trigger;
		pcrecpp::RE filter_stack_trigger_exception;
		
		std::list<std::list<std::string> > valuefilters;
};

INIT_MODULE(LoggerModule);
