/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>

#include <arpa/inet.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include "socket.hpp"

#include "buffer.hpp"

//consider using pcap
class SnifferModule : public SocketModuleBase
{
	public:
		SnifferModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			SocketModuleBase("Sniffer", modules, manager, configuration, database)
		{
			//TODO
			//Include the ability to specify interface(s) and whether the program
			//should explicitly enable, or disable promiscuous mode
			//Add a thing that specifies the interface in the sniff event
			
			//TODO: We might like to introduce the concept of a priority event that
			//gets run as soon as it's published, rather than being put in the queue,
			//and has the power to block any further processing of that event
			//This would be useful for, for example, IP address received from the wrong device
			//To avoid having to change event handlers to bool, perhaps have
			//GetManager()->BlockEvent(); that would block the current priority event, or
			//throw FormatError() if the currently processed event is not a priority event
			//Maybe change Subscribe(EventSubscription *) to Subscribe(EventSubscription *, bool = false)
			//With the bool being "priority"
			
			//TODO: copy more of these notes into the log
			
			//TODO: investigate PF_PACKET vs AF_PACKET
			sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
			if (sock < 0){
				throw FormatError("socket(): %s", errstr);
			}
			
			std::string name = (std::string)GetConfiguration()["sniffer"]["device"];
			if (setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, name.c_str(), name.size()) < 0){
				throw FormatError("setsockopt(): %s", errstr);
			}
			
			ifreq dev;
			strncpy(dev.ifr_name, name.c_str(), IFNAMSIZ);
			if (ioctl(sock, SIOCGIFINDEX, &dev) < 0){
				throw FormatError("ioctl(SIOCGIFINDEX): %s", errstr);
			}
			index = dev.ifr_ifindex;
			if (ioctl(sock, SIOCGIFHWADDR, &dev) < 0){
				throw FormatError("ioctl(SIOCGIFHWADDR): %s", errstr);
			}
			memcpy(addr, dev.ifr_hwaddr.sa_data, 6);
			
			AddSocket(sock);
			GetManager()->Subscribe(GetCallbackNameEventSubscription<SnifferModule>
				("send:sniff", this, &SnifferModule::OnSend));
		}
		
		//add a destructor to close the sock
		
		//this might need to be changed to support multiple sockets
		//(i.e. add a parameter (e.g. an int sock))
		virtual void HandleSocketEvent()
		{
			Buffer b;
			unsigned char *buffer = b.GetBuffer(65536);
			
			//TODO: maybe we can change this to get a sockaddr_ll that would
			//tell us the direction of the packet?
			int length = read(sock, buffer, 65536);
			if (length < 0){
				throw FormatError("read(): %s", errstr);
			}
			else if (!length){
				throw FormatError("read() returned zero");
			}
			b = Buffer(b, 0, length);
			
			Event e;
			e["received:ethernet"]["data"] = b;
			e["sniff"]["data"] = b;
			GetManager()->Publish(e);
		}
		
		void OnSend(Event event)
		{
			sockaddr_ll lladdr;
			lladdr.sll_family = AF_PACKET;
			lladdr.sll_protocol = 0;
			lladdr.sll_ifindex = index;
			lladdr.sll_hatype = 0;
			lladdr.sll_pkttype = 0;
			lladdr.sll_halen = 6;
			
			memcpy(lladdr.sll_addr, addr, 6);
			lladdr.sll_addr[6] = 0;
			lladdr.sll_addr[7] = 0;
			
			Buffer payload = event["send:sniff"]["payload"];
			if (sendto(sock, payload.GetBuffer(), payload.GetLength(), 0, (struct sockaddr*)&lladdr, sizeof(lladdr)) < 0){
				throw FormatError("write(): %s", errstr);
			}
		}
		
	private:
		unsigned char addr[6];
		int index;
		int sock;
		ifreq dev;
};

INIT_MODULE(SnifferModule);
