/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

//TODO: perhaps we should implement an accessor event?
//TODO: perhaps modules should be wrapped in namespaces?

#include "event.hpp"
#include "module.hpp"
#include "structures.hpp"

#include <cstdlib>
#include <list>
#include <map>
#include <pcrecpp.h>
#include <set>

//Because of syntax with <<>> this is the easiest way to do this:
class TracerouteAddressPair : public ComparisonPair<Buffer, Buffer>
{
	public:
		TracerouteAddressPair(Buffer pa, Buffer pb) :
			ComparisonPair<Buffer, Buffer>(pa, pb){}
		
	private:
};


struct TracerouteRequest
{
	TracerouteRequest()
	{
		ip = Buffer();
	}
	
	Time sent;
	Time received;
	uint16_t seq;
	DynamicValue ip;
};

class Traceroute
{
	public:
		Traceroute(){}
		
		Traceroute(Event pipv4, EventManager *pmanager, Database *pdb, unsigned int ptraceroute_id,
			unsigned int pinterval, unsigned int pmaxttl, unsigned int pn)
		{
			while (!current_id){
				//This probably isn't actually that random, but hopefully it'll do
				current_id = rand() ^ (rand() << 8);
			}
			
			interval = pinterval;
			maxttl = pmaxttl;
			n = pn;
			ipv4 = pipv4;
			manager = pmanager;
			db = pdb;
			traceroute_id = ptraceroute_id;
			
			reply = false;
			seq = 0;
			id = current_id;
			current_id++; //be ready for the next traceroute
			
			SendTimerEvent();
		}
		
		void OnEchoReply(Event ipv4event, Event icmpevent)
		{
			uint16_t sequence = icmpevent["processed:icmp:echo reply:echo reply"]["seq"];
			uint16_t identification = icmpevent["processed:icmp:echo reply:echo reply"]["id"];
			DynamicValue ip = ipv4event["processed:ipv4"]["source"];
			reply = ProcessIcmp(sequence, identification, ip);
		}
		
		void OnTtl(Event ipv4event, Event icmpevent)
		{
			//the payload of the IPv4 packet
			Buffer b = icmpevent["processed:icmp:time exceeded:ttl"]["payload"];
			if (b.GetLength() != 8){
				return; //TODO: something went wrong
			}
			
			uint16_t sequence = b.GetUInt16N(6);
			uint16_t identification = b.GetUInt16N(4);
			DynamicValue ip = ipv4event["processed:ipv4"]["source"];
			ProcessIcmp(sequence, identification, ip);
		}
			
		
		void OnTimer(Event event)
		{
			//we send the next packet in our traceroute
			SendTimerEvent();
		}
		
	private:
		bool ProcessIcmp(uint16_t sequence, uint16_t identification, DynamicValue ip)
		{
			if (id != identification){
				return false; //wrong traceroute or something
			}
			for (std::list<TracerouteRequest>::iterator it = requests.begin(); it != requests.end(); it++){
				//we found the right one:D
				if (it->seq == sequence){
					Time now = Time::now();
					if (now - it->sent > Time::ms(interval)){
						return false; //we processed it, but we also expired - ignore?
							//TODO: this is also strange
					}
					it->received = now;
					it->ip = ip;
					return true; //we wouldn't expect to see another matching one
						//(because that's how we constructed this)
				}
			}
			return false; //TODO: something very strange has happened - but we aren't checking it yet
		}
		
		void SendTimerEvent()
		{
			//Either we reached the maximum (we expect to send maxttl packets, so maxttl + 1)
			//or we received a reply, so should wait until the next ttl level
			//For more paranoia, check beyond this ttl (keep tracerouting) - but this isn't currently implemented
			
			if ((seq == (maxttl + 1) * n) || (reply && ((seq % n) == 0))){
				NewTracerouteCompleteEvent();
				return; //we've already sent all the packets we should
			}
			
			//Set the timeout
			Event e;
			e["traceroute:timer"]["destination"] = ipv4["send:ipv4"]["destination"];
			e["traceroute:timer"]["source"] = ipv4["send:ipv4"]["source"];
			Event t;
			t["timer"]["event"] = e;
			t["timer"]["time"] = Time::ms(interval);
			manager->Publish(t);
			
			//Create the packet to send
			unsigned int ttl = seq / n;
			Event icmp;
			Event ipv4i = ipv4;
			ipv4i["send:ipv4"]["ttl"] = ttl;
			icmp["send:icmp:echo request:echo request"]["seq"] = seq;
			icmp["send:icmp:echo request:echo request"]["id"] = id;
			icmp["send:icmp:echo request:echo request"]["payload"] = Buffer();
			icmp["send:icmp:echo request:echo request"]["next"] = ipv4i;
			icmp["send:icmp:echo request:echo request"]["next-name"] = "send:ipv4";
			manager->Publish(icmp);
			
			//Create the new request item with the current time
			TracerouteRequest request;
			request.sent = Time::now();
			request.received = Time::s(0);
			request.seq = seq;
			requests.push_back(request);
			seq++;
		}
		
		void NewTracerouteCompleteEvent()
		{
			std::vector<DynamicValue> l;
			for (std::list<TracerouteRequest>::iterator it = requests.begin(); it != requests.end(); it++){
				DynamicValue v;
				Buffer bip = it->ip;
				if (bip.GetLength()){
					double d = it->received - it->sent;
					v["address"] = it->ip;
					v["rtt"] = d;
					v["ttl"] = it->seq / n;
					l.push_back(v);
				}
			}
			Event e;
			e["traceroute:new traceroute"]["source"] = ipv4["send:ipv4"]["source"];
			e["traceroute:new traceroute"]["destination"] = ipv4["send:ipv4"]["destination"];
			e["traceroute:new traceroute"]["list"] = l;
			manager->Publish(e);
			
			for (std::list<TracerouteRequest>::iterator it = requests.begin(); it != requests.end(); it++){
				Buffer bip = it->ip;
				if (bip.GetLength()){
					DynamicValue v;
					v["traceroute"] = traceroute_id;
					v["n"] = it->seq % n;
					v["ttl"] = it->seq / n;
					v["rtt"] = (it->received - it->sent).GetNs();
					v["ip"] = bip;
					(*db)("INSERT INTO traceroute_response (traceroute, ttl, rtt, ip)\
						VALUES ($traceroute, $ttl, $rtt, $ip);")(v);
				}
			}
			
			//printf("TRACEROUTE: Should finish traceroute\n");
		}
		
		std::list<TracerouteRequest> requests;
		Event ipv4;
		unsigned int seq;
		bool reply;
		
		unsigned int interval;
		unsigned int maxttl;
		unsigned int n;
		uint16_t id;
		EventManager *manager;
		Database *db;
		unsigned int traceroute_id;
		
		static uint16_t current_id;
};

uint16_t Traceroute::current_id = 0;

struct TracerouteQueueItem
{
	DynamicValue src;
	DynamicValue dst;
	std::string direction;
};

class TracerouteModule : public Module
{
	public:
		TracerouteModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Traceroute", modules, manager, configuration, database),
			filter_source(GetConfiguration()["traceroute"]["source_ips"]),
			filter_destination(GetConfiguration()["traceroute"]["destination_ips"])
		{
			std::list<std::string> ipv4echoreply;
			ipv4echoreply.push_back("processed:ipv4");
			ipv4echoreply.push_back("processed:icmp:echo reply:echo reply");
			GetManager()->Subscribe(GetNetworkStackEventSubscription<TracerouteModule>
				(ipv4echoreply, this, &TracerouteModule::OnEchoReply));
			
			std::list<std::string> ipv4timeexceeded;
			ipv4timeexceeded.push_back("processed:ipv4");
			ipv4timeexceeded.push_back("processed:icmp:time exceeded:ttl");
			GetManager()->Subscribe(GetNetworkStackEventSubscription<TracerouteModule>
				(ipv4timeexceeded, this, &TracerouteModule::OnTtl));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<TracerouteModule>
				("traceroute:timer", this, &TracerouteModule::OnTimer));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<TracerouteModule>
				("processed:ipv4", this, &TracerouteModule::OnIpv4));
			
			//This seems easier than passing parents and doing this in direct code
			//(although we could do that too)
			GetManager()->Subscribe(GetCallbackNameEventSubscription<TracerouteModule>
				("traceroute:new traceroute", this, &TracerouteModule::OnTracerouteComplete));
			
			GetDatabase()("CREATE TABLE IF NOT EXISTS traceroute_traceroute\
				(id INTEGER PRIMARY KEY,\
				src BINARY NOT NULL,\
				dst BINARY NOT NULL, \
				timestamp INTEGER NOT NULL);")();
			
			GetDatabase()("CREATE TABLE IF NOT EXISTS traceroute_response\
				(traceroute INTEGER NOT NULL,\
				ttl INTEGER NOT NULL,\
				rtt INTEGER NOT NULL,\
				ip BINARY NOT NULL,\
				FOREIGN KEY(traceroute) REFERENCES traceroute_traceroute(id));")();
			
			DynamicValue v = GetDatabase()("SELECT src,dst,timestamp FROM traceroute_traceroute;")();
			std::vector<DynamicValue> vv = v;
			for (std::vector<DynamicValue>::iterator it = vv.begin(); it != vv.end(); it++){
				TracerouteAddressPair addrpair = TracerouteAddressPair((*it)["src"], (*it)["dst"]);
				dbtraceroutes[addrpair] = Time::ns((*it)["timestamp"]);
			}
		}
		
	private:
		
		void OnIpv4(Event event)
		{
			//make this add to a queue? (or just traceroute all intermediate nodes too)
			//this means we have an O(n^2) traceroute, but it does also mean if the
			//traceroute goes a different way depending on IP address after the intermediate
			//node, then this would be detected
			//Actually, this system is not as bad as O(n^2) as repeated routers will be
			//tracerouted only once. Indeed, if all 4 billion IPv4 addresses are reached,
			//it will reduce to O(n)
			
			//TODO: Explain in the dissertation why:
			//a) This property of separately tracerouting intermediate nodes is good
			//b) Why this probably won't degenerate into the O(n^2) case too much
			
			//TODO: Implement some security to prevent us from being asked to traceroute the whole of:
			//the IP address space because some node is spamming us with lots of different sourced packets
			//(either that, or have a threshold above which we have an anomaly for too many IPs?)
			//This will be particularly relevant to IPv6, where we don't have enough memory to store
			//2^128 lots of data.
			
			bool traceicmp = GetConfiguration()["traceroute"]["traceicmp"];
			unsigned int proto = event["processed:ipv4"]["protocol"];
			if (!traceicmp && (1 == proto)){
				//This is an ICMP packet, and we aren't supposed to traceroute those
				return;
			}
			
			//perform a traceroute from src to dst (intended for if localhost sends a packet)
			TryTraceroute(event["processed:ipv4"]["source"], event["processed:ipv4"]["destination"], "destination");
			
			//perform a traceroute from dst to src (intended for if localhost receives a packet)
			TryTraceroute(event["processed:ipv4"]["destination"], event["processed:ipv4"]["source"], "source");
		}
		
		void OnEchoReply(std::vector<Event> &events)
		{
			TracerouteAddressPair addrs(events[0]["processed:ipv4"]["destination"], events[0]["processed:ipv4"]["source"]);
			if (traceroutes.count(addrs)){
				traceroutes[addrs].OnEchoReply(events[0], events[1]);
			}
		}
		
		void OnTtl(std::vector<Event> &events)
		{
			//not an ICMP packet that caused this
			unsigned int protocol = events[1]["processed:icmp:time exceeded:ttl"]["protocol"]; //TODO: icmp time exceeded
				//TODO: actually sending packets!
			if (protocol != 1){
				return;
			}
			TracerouteAddressPair addrs(
				events[1]["processed:icmp:time exceeded:ttl"]["source"],
				events[1]["processed:icmp:time exceeded:ttl"]["destination"]);
			//TODO: validate that the source and destination make sense
			//(i.e. guard against spooding)
			if (traceroutes.count(addrs)){
				traceroutes[addrs].OnTtl(events[0], events[1]);
			}
		}
		
		void OnTimer(Event event)
		{
			TracerouteAddressPair addrs(event["traceroute:timer"]["source"], event["traceroute:timer"]["destination"]);
			if (traceroutes.count(addrs)){
				traceroutes[addrs].OnTimer(event);
			}
		}
		
		//Perform a traceroute from src to dst
		//BUG: Buffer, std::string - needs to be DynamicValue
		void TryTraceroute(DynamicValue src, DynamicValue dst, std::string direction)
		{
			TracerouteAddressPair addrs(src, dst);
			
			if (traceroutes.count(addrs) || dbtraceroutes.count(addrs) || traceroute_queue_set.count(addrs)){
				//printf("Already in\n");
				return; //We already did the traceroute
				//TODO: make this be dependent on the time
				//(we really want to rerun a traceroute rather than destroy and rerun)
			}
			
			if (!filter_source.FullMatch(src.ToString()) || !filter_destination.FullMatch(dst.ToString()))
			{
				//printf("source or destination mismatch\n");
				return; //Nothing special here:p
			}
			//if we get here, we're configured to traceroute
			
			
			TracerouteQueueItem item;
			item.src = src;
			item.dst = dst;
			item.direction = direction;
			traceroute_queue.push_back(item); //Add the address to the queue
			traceroute_queue_set.insert(addrs); //Add the address to the queue
				//set
			
			//TODO: Change this to < n (where n is configurable)
			//TODO: Check for too many items in the queue (SECURITY)
			//printf("TRACEROUTE: src: %s, dst: %s    ", src.ToString().c_str(), dst.ToString().c_str());
			//printf("TRACEROUTE: %u (%u)\n", (unsigned int)traceroute_queue.size(), (unsigned int)traceroutes.size());
			if (traceroute_queue.size() == 1){
				//There isn't another
				AddTraceroute();
			}
		}
		
		void OnTracerouteComplete(Event event)
		{
			//NOTE: USEFUL PRINTF
			//printf("FINISHED TRACEROUTE (%u remaining)\n", (unsigned int)traceroute_queue.size());
			//TODO: this logic will need to change when we make this < n
			traceroute_queue.pop_front(); //Until then, the front must be the
				//one that just finished
			if (traceroute_queue.size()){
				//we have at least one traceroute to do, so do it
				AddTraceroute();
			}
		}
		
		void AddTraceroute()
		{
			TracerouteQueueItem item = traceroute_queue.front();
			
			DynamicValue src = item.src;
			DynamicValue dst = item.dst;
			std::string direction = item.direction;
			TracerouteAddressPair addrs(src, dst);
			traceroute_queue_set.erase(addrs); //remove the address from the
				//queue set
			
			//NOTE: USEFUL PRINTF
			//printf("TRACEROUTE: ******** %s -> %s ********\n", src.ToString().c_str(), dst.ToString().c_str());
			
			Event e;
			e["traceroute:new ipv4"]["source"] = src;
			e["traceroute:new ipv4"]["destination"] = dst;
			e["traceroute:new ipv4"]["direction"] = direction;
			GetManager()->Publish(e);
			
			Event ipv4;
			
			
			//NOTE: to move to log - we might like to note somewhere that the traceroute
			//class is a good test of SENDING stuff
			//TODO: At some point run valgrind on this program
			//TODO: Improve the FSM created by GetNetworkStackEventSubscription()
			//to handle more corner cases
			
			//This send:auto works out both the destination and source mac address
			//based on the destination and source IP (you wouldn't expect, for example,
			//that the source IP is that of another machine)
			//an event that is send:ipv4 and send:auto would
				//probably be handled by the ARP module, and moved up to ethernet
				//and then to sniff (possibly by producing a send:ethernet;send:auto)
				//it would therefore not generally be a good idea to implement
				//things such that more than one send:auto event handler will
				//cause physical traffic to be produced
				//NOTE: this is probably NOT incompatible with the idea of having
				//a protocol or module tree
			//Actually, that won't work because then send:ipv4 will intercept it
			//(and we won't get the buffer for it)
			
			Event sendauto;
			sendauto["send:auto:ipv4"];
			ipv4["send:ipv4"]["next"] = sendauto; //this might work
			ipv4["send:ipv4"]["next-name"] = "send:auto:ipv4";
				//we then have ipv4 module send with:
				//["send:auto:ipv4"]["data"]
				//["send:auto:ipv4"]["ipv4"] = the stuff from ["send:ipv4"]
			ipv4["send:ipv4"]["protocol"] = 1; //ICMP
			ipv4["send:ipv4"]["source"] = src;
			ipv4["send:ipv4"]["destination"] = dst;
			//it's intended that ttl, protocol, and payload are set elsewhere
			
			//TODO: This will have to change when we allow these to time out
			
			DynamicValue dbtraceroute;
			dbtraceroute["src"] = src;
			dbtraceroute["dst"] = dst;
			dbtraceroute["timestamp"] = Time::now().GetNs();
			GetDatabase()("INSERT INTO traceroute_traceroute\
				(src, dst, timestamp) VALUES ($src, $dst, $timestamp);")(dbtraceroute);
				
			DynamicValue tnrow = GetDatabase()("SELECT id FROM traceroute_traceroute WHERE\
				src=$src AND dst=$dst AND timestamp=$timestamp;")(dbtraceroute);
			
			traceroutes[addrs] = Traceroute(ipv4, GetManager(), &GetDatabase(), tnrow[(unsigned int)0]["id"],
				GetConfiguration()["traceroute"]["interval"],
				GetConfiguration()["traceroute"]["maxttl"],
   				GetConfiguration()["traceroute"]["n"]);
		}
		
		std::map<TracerouteAddressPair, Traceroute> traceroutes;
		std::map<TracerouteAddressPair, Time> dbtraceroutes;
		
		std::list<TracerouteQueueItem> traceroute_queue;
		std::set<TracerouteAddressPair> traceroute_queue_set;
		
		pcrecpp::RE filter_source;
		pcrecpp::RE filter_destination;
};

INIT_MODULE(TracerouteModule);
