/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include "event.hpp"
#include "module.hpp"

class UdpModule : public Module
{
	public:
		UdpModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("UDP", modules, manager, configuration, database)
		{
			std::list<std::string> ipv4udp;
			ipv4udp.push_back("processed:ipv4");
			ipv4udp.push_back("received:udp");
			GetManager()->Subscribe(GetNetworkStackEventSubscription<UdpModule>
				(ipv4udp, this, &UdpModule::OnReceiveIpv4));
		}
		
	private:
		void OnReceiveIpv4(std::vector<Event> &events)
		{
			OnReceive("ipv4", events[0]["processed:ipv4"]["source"], events[0]["processed:ipv4"]["destination"],
				events[1]["received:udp"]["data"]);
		}
		
		void OnReceive(std::string netname, DynamicValue src, DynamicValue dst, Buffer b)
		{
			DynamicValue ev;
			ev["network name"] = netname;
			ev["source address"] = src;
			ev["destination address"] = dst;
			ev["data"] = b;
			
			if (b.GetLength() < 8){
				PublishEvent("anomaly:received:udp:packet length < header length", ev);
				return;
			}
			
			uint16_t sport = b.GetUInt16N(0);
			uint16_t dport = b.GetUInt16N(2);
			uint16_t length = b.GetUInt16N(4);
			
			ev["source port"] = sport;
			ev["destination port"] = dport;
			
			/*TODO: Checksum*/
			if (b.GetLength() < length){
				ev["payload length"] = length;
				ev["received length"] = length;
				PublishEvent("anomaly:received:udp:incomplete packet", ev);
				return;
			}
			else if (length < 8){
				ev["payload length"] = length;
				ev["received length"] = length;
				PublishEvent("anomaly:received:udp:packet too small", ev);
				return;
			}
			
			ev["payload"] = Buffer(b, 8);
			PublishEvent("processed:udp", ev);
		}
		
		void PublishEvent(std::string name, DynamicValue ev)
		{
			Event e;
			e[name] = ev;
			GetManager()->Publish(e);
		}
};

INIT_MODULE(UdpModule);
