/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <netinet/in.h>

#include "socket.hpp"

SocketModuleBase::SocketModuleBase(const char *pname, Modules *modules, EventManager *manager,
	Configuration *configuration, Database *database) :
	Module(pname, modules, manager, configuration, database)
{
	GetManager()->Subscribe(GetCallbackNameEventSubscription<SocketModuleBase>
		("socket", this, &SocketModuleBase::OnSocketEvent));
	//GetManager()->Subscribe(
	//	new TypeCallbackEventSubscription<SocketModuleBase, SocketEvent>(this, &SocketModuleBase::OnSocketEvent));
}

void SocketModuleBase::OnSocketEvent(Event event)
{
	HandleSocketEvent();
}

void SocketModuleBase::AddSocket(int socket)
{
	Event e;
	DynamicValue v;
	v["file descriptor"] = socket;
	v["this"] = this;
	e["socket:new"] = v;
	GetManager()->Publish(e);
	//GetManager()->Publish(new NewSocketEvent(socket, this));
}

void SocketModuleBase::RemoveSocket(int socket)
{
	throw FormatError("Not implemented yet");
}

SocketModule::SocketModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
	Module("Socket", modules, manager, configuration, database)
{
	GetManager()->Subscribe(GetCallbackNameEventSubscription<SocketModule>
		("socket:new", this, &SocketModule::OnNewSocket));
	//GetManager()->Subscribe(
	//	new TypeCallbackEventSubscription<SocketModule, NewSocketEvent>(this, &SocketModule::OnNewSocket));
}

//I have a very similar function in another program I've written, so some of this is from there
void SocketModule::ProcessSockets(Time t)
{
	//avoid waiting indefinitely
	if (!t && !sockets.size()){
		throw FormatError("No sockets and no timers!");
	}
	
	//set up the fd set
	fd_set fdset;
	FD_ZERO(&fdset); //zero out the file descriptor set
	//add each socket in the map to the set
	for (std::map<int, SocketModuleBase *>::iterator it = sockets.begin(); it != sockets.end(); it++){
		FD_SET((*it).first, &fdset); //does the actual adding
	}
	
	if (t != Time::s(0)){
		/*struct timeval timeout; //the time structure
		timeout.tv_sec = t/1000000; //seconds
		timeout.tv_usec = t%1000000; //microseconds*/
		//wait for the timeout, or a socket to be ready to receive data
		struct timeval timeout = t;
		/*
		unsigned int a = timeout.tv_sec;
		unsigned int b = timeout.tv_usec;
		
		printf("WAITING: %u %u\n", a, b);*/
		if (select(FD_SETSIZE, &fdset, 0, 0, &timeout) < 0){ //error
			throw FormatError("select(): %s", errstr);
		}
	}
	else {
		if (select(FD_SETSIZE, &fdset, 0, 0, 0) < 0){ //error
			throw FormatError("select(): %s", errstr);
		}
	}
	
	for (int i = 0; i < FD_SETSIZE; i++){
		if (FD_ISSET(i, &fdset)){ //this socket received data
			try {
				if (!sockets.count(i)){
					throw FormatError("Received from socket not placed in FD_SET");
				};
				
				try
				{
					sockets[i]->HandleSocketEvent();//call the read method on it
				}
				catch (Error e)
				{
					e.OutputMessage();
				}
			}
			catch (Error e) {
				e.OutputMessage(); //for some reason this doesn't always get called
				//delete sockets[i]; //something went wrong with the socket, so delete it
					//(avoid infinite loops of messages)
				sockets.erase(i); //instead of delete, remove from the list
			}
			catch (std::exception& e) {
				fprintf(stderr, "C++ Exception: %s", e.what());
				//delete sockets[i]; //something went wrong with the socket, so delete it
					//(avoid infinite loops of messages)
				sockets.erase(i); //instead of delete, remove from the list
			}
			return; //we need to leave now to preven weird bugs
		}
	}
}

void SocketModule::OnNewSocket(Event event)
{
	SocketModuleBase *sm = (SocketModuleBase *)((void *)event["socket:new"]["this"]);
	int sock = event["socket:new"]["file descriptor"];
	if (sockets.count(sock)){
		throw FormatError("Socket already registered!");
	}
	sockets[sock] = sm;
}