/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011-2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

//I wouldn't expect this header to be included directly
#ifndef EVENT_HPP
#error Include event.hpp, not subscriptions.hpp
#endif

#include "config.h"
#include "error.hpp"

#ifndef SUBSCRIPTIONS_HPP
#define SUBSCRIPTIONS_HPP

#include <list>
#include <set>
#include <string>
#include <vector>

#include "fsm.hpp"

class FsmEventSubscriptionNode : public FsmNode<Event &, std::vector<Event> >
{
	public:
	private:
};

class FsmEventSubscriptionFsm : public Fsm<Event &, std::vector<Event> >
{
	public:
		FsmEventSubscriptionFsm(FsmNode<Event &, std::vector<Event> > *first) :
			Fsm<Event &, std::vector<Event> >(std::vector<Event>(), first){}
		
	private:
};

class FsmEventSubscription : public EventSubscription
{
	public:
		FsmEventSubscription(FsmEventSubscriptionNode *first) :
			fsm(first)
		{
			nodes.insert(first);
		}
		
		~FsmEventSubscription()
		{
			for (std::set<FsmEventSubscriptionNode *>::iterator it = nodes.begin(); it != nodes.end(); it++){
				delete *it;
			}
		}
		
		void Add(FsmEventSubscriptionNode *node)
		{
			nodes.insert(node);
		}
		
		virtual void HandleEvent(Event &event)
		{
			fsm.ProcessCharacter(event);
		}
		
	private:
		FsmEventSubscriptionFsm fsm;
		std::set<FsmEventSubscriptionNode *> nodes;
};


class NameEventSubscription : public EventSubscription
{
	public:
		NameEventSubscription(std::string pname, EventSubscription *pnext)
		{
			name = pname;
			next = pnext;
		}
		
		~NameEventSubscription()
		{
			delete next;
		}
		
		virtual void HandleEvent(Event &event)
		{
			if (event.Has(name)){
				next->HandleEvent(event);
			}
		}
		
	private:
		std::string name;
		EventSubscription *next;
};

template <typename T> class NameValueEventSubscription : public EventSubscription
{
	public:
		NameValueEventSubscription(std::string pname, std::string pkey, T pvalue, EventSubscription *pnext)
		{
			name = pname;
			key = pkey;
			value = pvalue;
			next = pnext;
		}
		
		~NameValueEventSubscription()
		{
			delete next;
		}
		
		virtual void HandleEvent(Event &event)
		{
			if (event.Has(name) && event[name].Has(key)){
				Error::InvertCrash();
				try {
					T v = event[name][key];
					Error::InvertCrash();
					if (v == value){
						next->HandleEvent(event);
					}
				}
				catch (Error e){
					Error::InvertCrash();
					return;
				}
			}
		}
		
	private:
		std::string name;
		std::string key;
		T value;
		EventSubscription *next;
};

template <typename H> class CallbackEventSubscription : public EventSubscription
{
	public:
		CallbackEventSubscription(H *pobject, void (H::*pfunction)(Event))
		{
			object = pobject;
			function = pfunction;
		}
		
		virtual void HandleEvent(Event &event)
		{
			((object)->*(function))(event);
		}
		
	private:
		H *object;
		void (H::*function)(Event);
};

template <typename H> struct NameFsmEventSubscriptionMatch
{
	std::string name;
	FsmEventSubscriptionNode *node;
	void (H::*function)(std::vector<Event> &);
	bool append;
	bool clear;
};

template <typename H> class NameFsmEventSubscriptionNode : public FsmEventSubscriptionNode
{
	public:
		NameFsmEventSubscriptionNode(H *pobject)
		{
			mismatch = 0;
			mismatchfn = 0;
			object = pobject;
			mismatchappend = false;
			mismatchclear = false;
		}
		
		//Note: the FIRST match is the one that's used
		//e.g. if we have an event with sniff, and received:ethernet
		//and a match rule for sniff and received:ethernet,
		//whichever was added first will be used
		virtual FsmEventSubscriptionNode *ProcessCharacter(Event &event, std::vector<Event> &events)
		{
			//I wonder what the typename part does
			for (typename std::list<NameFsmEventSubscriptionMatch<H> >::iterator it = matches.begin();
				it != matches.end(); it++){
				if (event.Has(it->name)){
					if (it->append){
						events.push_back(event);
					}
					if (it->function){
						((object)->*(it->function))(events);
					}
					if (it->clear){
						events.clear();
					}
					return it->node;
				}
			}
			
			if (mismatchclear){
				events.clear();
			}
			if (mismatchfn){
				((object)->*(mismatchfn))(events);
			}
			if (mismatchappend){
				events.push_back(event);
			}
			
			return mismatch;
		}
		
		void SetMatch(std::string name, FsmEventSubscriptionNode *node, bool clear = false, bool append = false,
			void (H::*function)(std::vector<Event> &) = 0)
		{
			NameFsmEventSubscriptionMatch<H> match;
			match.name = name;
			match.node = node;
			match.function = function;
			match.append = append;
			match.clear = clear;
			matches.push_back(match);
		}
		
		void SetMismatch(FsmEventSubscriptionNode *node,
			bool clear = false, bool append = false, void (H::*function)(std::vector<Event> &) = 0)
		{
			mismatch = node;
			mismatchfn = function;
			mismatchappend = append;
			mismatchclear = clear;
		}
		
	private:
		std::list<NameFsmEventSubscriptionMatch<H> > matches;
		FsmEventSubscriptionNode *mismatch;
		void (H::*mismatchfn)(std::vector<Event> &);
		H *object;
		bool mismatchappend;
		bool mismatchclear;
};

/*Convenience functions*/
template <typename H> EventSubscription *GetCallbackNameEventSubscription
	(std::string name, H *pobject, void (H::*pfunction)(Event))
{
	return new NameEventSubscription(name, new CallbackEventSubscription<H>(pobject, pfunction));
}

template <typename H, typename T> EventSubscription *GetCallbackNameValueEventSubscription
	(std::string name, std::string key, T value, H *pobject, void (H::*pfunction)(Event))
{
	return new NameValueEventSubscription<T>(name, key, value, new CallbackEventSubscription<H>(pobject, pfunction));
}

template <typename H> EventSubscription *GetNetworkStackEventSubscription
	(std::list<std::string> names, H *object, void (H::*function)(std::vector<Event> &))
{
	//Check that we have at least two
	if (!names.size()){
		throw FormatError("No names in stack");
	}
	else if (names.size() == 1){
		throw FormatError("Only one name in stack");
	}
	
	NameFsmEventSubscriptionNode<H> *first = new NameFsmEventSubscriptionNode<H>(object);
	NameFsmEventSubscriptionNode<H> *last = first;
	FsmEventSubscription *fsm = new FsmEventSubscription(first);
	
	std::list<std::string>::iterator it = names.begin();
	while (true){
		std::string name = *it;
		it++;
		
		last->SetMatch("sniff", first, true);
		last->SetMismatch(last);
		fsm->Add(last);
		
		if (it == names.end()){
			last->SetMatch(name, first, true, true, function);
			break;
		}
		else {
			NameFsmEventSubscriptionNode<H> *node = new NameFsmEventSubscriptionNode<H>(object);
			last->SetMatch(name, node, false, true);
			last = node;
		}
	}
	
	return fsm;
}

#endif
